@@ -1,4 +1,4 @@
-/* Source Installer, Copyright (c) 2005 Claudio Fontana
+/* Source Installer, Copyright (c) 2005-2006 Claudio Fontana
 
  actions.c - actions implementation
 
@@ -26,102 +26,36 @@
 #include "actions.h"
 
 static int _configure(char** cflags);
-/* static int _make_clean(char* makefile); */
 static int _make_uninstall(struct _package_info* info, char* makefile);
 static int _internal_uninstall(struct _package_info* info);
-static int _get_relative_dependencies(struct srcinst_string_list* dep, struct _package_info* info);
+static int _get_relative_dependencies(struct srcinst_string_list* dep, struct _package_info* info, int mirror);
 static int _count_slashes(char* path);
 static int _compare_dir_depth(char** path1, char** path2);
 static int _compare_strcmp(char** str1, char** str2);
+static void _strip_dirfile(struct srcinst_string_list* list);
 
 static int _make(char* makefile);
-static int _make_install(struct _package_info* info, char* makefile, int strip, int force);
-static char* _get_makefile(void);
+static SRCINST_ERR _make_test_install(struct _package_info* info, struct _makefile_features* mf, int strip);
+static SRCINST_ERR _make_install(struct _package_info* info, struct _makefile_features* mf, int strip);
 static char* _keep_source(struct _package_info* info, SRCINST_COMP format);
 
 static char* _unpack(char* filename);
 static int _delete_source(char* filename);
 static int _is_valid_package(char* filename);
 
-static int _find_files_aux(struct srcinst_string_list* files, char* dirname, char** good_files);
-
 static char* _access_string(char** set, int i);
 
-static void _init_makefile_features(struct _makefile_features* mf);
-static int _test_makefile_features(struct _makefile_features* mf, char* mfile);
 
-static int _test_makefile_features_rec(struct _makefile_features* mf, char* mfile);
+/* detect makefile features, and return them as an opaque type.
+   Note that the structure also contains the makefile name */
 
-static void _init_makefile_features(struct _makefile_features* mf) {
-  mf->features[SRCINST_FEATURE_DESTDIR].pattern = "\\$[({]?DESTDIR[)}]?";
-  mf->features[SRCINST_FEATURE_DESTDIR].supported = 0;
-  mf->features[SRCINST_FEATURE_INSTALL_ROOT].pattern = "\\$[({]?INSTALL_ROOT[)}]?";
-  mf->features[SRCINST_FEATURE_INSTALL_ROOT].supported = 0;
-  mf->features[SRCINST_FEATURE_UNINSTALL].pattern = "uninstall *:";
-  mf->features[SRCINST_FEATURE_UNINSTALL].supported = 0;
-  mf->features[SRCINST_FEATURE_STRIP].pattern = "install-strip *:";
-  mf->features[SRCINST_FEATURE_STRIP].supported = 0;
-}
+struct _makefile_features* _action_detect_makefile_features(struct _package_info* i) {
 
-static int _test_makefile_features(struct _makefile_features* mf, char* mfile) {
-  int i; FILE* m; off_t m_size;
-  char* buffer;
-  
-  _init_makefile_features(mf);
-
-  if ((m_size = srcinst_file_size(mfile)) == (off_t) -1) {
-    return 0;
-  }
-
-  buffer = srcinst_malloc(m_size + 1);
-
-  if (!(m = fopen(mfile, "r")))
-    { free(buffer); return 0; }
-    
-  if ((fread(buffer, 1, m_size, m)) != m_size)
-    { free(buffer); fclose(m); return 0; }
-
-  fclose(m); buffer[m_size] = 0;
-
-  for (i = 0; i < SRCINST_FEATURE_N; i++) {
-    SRCINST_REGEX preg;
-    preg = srcinst_init_regex(mf->features[i].pattern, -1);
-    mf->features[i].supported = srcinst_exec_regex(preg, buffer, 0);
-    srcinst_free_regex(preg);
-  }
-  
-  free(buffer);
-  return 1;
-}
-
-static int _test_makefile_features_rec(struct _makefile_features* mf, char* makefile) {
-
-  if (!(_test_makefile_features(mf, makefile)))
+  if (_check_build_ready(i) != SRCINST_ERR_OK)
     return 0;
 
-  if (!mf->features[SRCINST_FEATURE_DESTDIR].supported &&
-      !mf->features[SRCINST_FEATURE_INSTALL_ROOT].supported) {
-    struct srcinst_string_list makefiles;
-    char* makefile_names[2];
-
-    srcinst_init_string_list(&makefiles);
-    makefile_names[0] = makefile; makefile_names[1] = 0;
-    
-    if (_find_files_aux(&makefiles, ".", makefile_names)) {
-      struct _makefile_features mf2; int i;
-
-      for (i = 0; i < makefiles.count; i++) {
-	if (!_test_makefile_features(&mf2, makefiles.strings[i]))
-	  break;
-	if (mf2.features[SRCINST_FEATURE_DESTDIR].supported)
-	  { mf->features[SRCINST_FEATURE_DESTDIR].supported = 1; break; }
-	if (mf2.features[SRCINST_FEATURE_INSTALL_ROOT].supported)
-	  { mf->features[SRCINST_FEATURE_INSTALL_ROOT].supported = 1; break; }
-      }
-    }
-    srcinst_free_string_list(&makefiles);
-  }
-  return 1;
+  return
+    _detect_makefile_features();
 }
 
 /* prepare source in the build directory for later use.
@@ -179,11 +113,11 @@ SRCINST_ERR _action_init_build(struct _package_info* info, char* filename, char*
   }
   
   if (!info->name) {
-  /* set default package name, ensuring uniqueness */
+  /* set default package name, verifying uniqueness */
 
     info->name = srcinst_strdup(info->packdir);
-    while (_search_package_list(&_srcinst_state.packages, info->name)) {
-      info->name = srcinst_append(info->name, "_new");
+    if (_search_package_list(&_srcinst_state.packages, info->name)) {
+      rv = SRCINST_ERR_BUSY; goto end_proc;
     }
   }
   
@@ -201,7 +135,7 @@ SRCINST_ERR _action_init_build(struct _package_info* info, char* filename, char*
   if (info->build_subdir && chdir(info->build_subdir) != 0) {
     rv = SRCINST_ERR_ACTION; goto end_proc;
   }
-  
+
   rv = SRCINST_ERR_OK;
   info->build_ready = 1;
   
@@ -219,8 +153,8 @@ SRCINST_ERR _action_init_build(struct _package_info* info, char* filename, char*
 SRCINST_ERR _action_chdir_build(struct _package_info* info, char* subdir) {
   SRCINST_ERR rv; char* target;
 
-  if (!_check_build_ready(info))
-    return SRCINST_ERR_INVALID;
+  if ((rv = _check_build_ready(info)) != SRCINST_ERR_OK)
+    return rv;
   
   target = srcinst_strjoin(_srcinst_state.dirs[SRCINST_DIR_BUILD], "/",
 			   info->packdir, "/", subdir ? subdir : "", 0);
@@ -254,41 +188,48 @@ SRCINST_ERR _action_fini_build(struct _package_info* info) {
   if (info->packdir) {
     free(info->packdir); info->packdir = 0;
   }
+
   info->build_ready = 0;
 
   return SRCINST_ERR_OK;
 }
 
-SRCINST_ERR _action_install_binary(struct _package_info* info, int strip, int force) {
-  char* makefile; SRCINST_ERR rv;
+SRCINST_ERR _action_build(struct _package_info* info, struct _makefile_features* mf) {
+  SRCINST_ERR rv;
+
+  if ((rv = _check_build_ready(info)) != SRCINST_ERR_OK)
+    return rv;
+
+  if (!_make(mf->makefile))
+    return SRCINST_ERR_ACTION;
   
-  if (!_check_build_ready(info))
-    return SRCINST_ERR_INVALID;
+  return SRCINST_ERR_OK;
+}
 
-  if (!(makefile = _get_makefile())) {
-    return SRCINST_ERR_MISSING;
-  }
+SRCINST_ERR _action_test_install(struct _package_info* info, struct _makefile_features* mf, int strip) {
+  SRCINST_ERR rv;
 
-  if (!_make(makefile)) {
-    rv = SRCINST_ERR_ACTION; goto end_proc;
-  }
+  if ((rv = _check_build_ready(info)) != SRCINST_ERR_OK)
+    return rv;
 
-  if (!_make_install(info, makefile, strip, force)) {
-    rv = SRCINST_ERR_ACTION; goto end_proc;
-  }
+  return _make_test_install(info, mf, strip);
+}
 
-  rv = SRCINST_ERR_OK;
+SRCINST_ERR _action_install_binary(struct _package_info* info, struct _makefile_features* mf, int strip) {
+  SRCINST_ERR rv;
 
- end_proc:
-  free(makefile);
-  return rv;
+  if ((rv = _check_build_ready(info)) != SRCINST_ERR_OK)
+    return rv;
+
+  return _make_install(info, mf, strip);
 }
 
 SRCINST_ERR _action_install_source(struct _package_info* info, SRCINST_COMP format) {
   char* previous_source, *tmp;
+  SRCINST_ERR rv;
 
-  if (!_check_build_ready(info))
-    return SRCINST_ERR_INVALID;
+  if ((rv = _check_build_ready(info)) != SRCINST_ERR_OK)
+    return rv;
 
   previous_source = info->source_location;
   if (!(tmp = _keep_source(info, format)))
@@ -309,6 +250,30 @@ SRCINST_ERR _action_install_source(struct _package_info* info, SRCINST_COMP form
   return SRCINST_ERR_OK;
 }
 
+int _action_get_conflict_list(struct _package_info* info,
+			      struct srcinst_string_list* list) {
+  int i;
+
+  /* get a deep copy of info dependencies */
+  if (!_get_relative_dependencies(list, info, 0))
+    return 0;
+  
+  for (i = 0; i < list->count; i++) {
+    int ft; size_t len;
+    ft = srcinst_file_type(list->strings[i]);
+
+    if (ft == SRCINST_TYPE_DIR ||
+
+	(len = strlen(list->strings[i])) >= 9 &&
+	strcmp(list->strings[i] + (len - 9), "/info/dir") == 0)
+
+      list->strings[i][0] = '\0';
+  }
+
+  srcinst_purge_string_list(list, 0);
+  return 1;
+}
+
 static char* _keep_source(struct _package_info* info, SRCINST_COMP fmt) {
   char* source_location;
   char* tarname, *zname, *cmd;
@@ -335,7 +300,7 @@ static char* _keep_source(struct _package_info* info, SRCINST_COMP fmt) {
 
   zname = srcinst_strjoin(tarname, srcinst_compression_format_to_ext(fmt), 0);
 
-  if (!srcinst_spawn_wait(PRG_mv, zname, _srcinst_state.dirs[SRCINST_DIR_SRC], 0))
+  if (!srcinst_spawn_wait(PRG_mv, "-f", zname, _srcinst_state.dirs[SRCINST_DIR_SRC], 0))
     goto end_proc;
 
   if (chdir(info->packdir) != 0)
@@ -361,11 +326,12 @@ SRCINST_ERR _action_uninstall_binary(struct _package_info* info) {
   if (!info->installed)
     return SRCINST_ERR_INVALID;
   
-  if (info->source_location) {
+  if (info->source_location && info->installed_files.count == 0) {
     char* makefile;
+    SRCINST_ERR tmprv;
 
-    if (!_check_build_ready(info))
-      return SRCINST_ERR_INVALID;
+    if ((tmprv = _check_build_ready(info)) != SRCINST_ERR_OK)
+      return tmprv;
   
     if (!(makefile = _get_makefile()))
       rv = 0;
@@ -374,7 +340,7 @@ SRCINST_ERR _action_uninstall_binary(struct _package_info* info) {
       free(makefile);
     }
   }
-
+  
   if (!rv && _srcinst_state.sigint_seen) {
     return SRCINST_ERR_INTR;
   }
@@ -403,16 +369,21 @@ SRCINST_ERR _action_uninstall_binary(struct _package_info* info) {
 }
 
 /* warning: DEP string_list gets the (char**) ->strings allocated,
-   but the single strings are mirrored from INFO. */
+   but the single strings are mirrored from info->installed_files */
 
 static int _get_relative_dependencies(struct srcinst_string_list* dep,
-				      struct _package_info* info) {
+				      struct _package_info* info, int mirror) {
+
   int i; size_t n;
   struct _package_list* packages;
   n = info->installed_files.count;
   
   packages = &_srcinst_state.packages;
-  srcinst_mirror_string_list(dep, &(info->installed_files));
+  
+  if (mirror)
+    srcinst_mirror_string_list(dep, &info->installed_files);
+  else
+    srcinst_copy_string_list(dep, &info->installed_files);
   
   for (i = 0; i < dep->count; i++) {
     SRCINST_TYPE ft;
@@ -422,6 +393,8 @@ static int _get_relative_dependencies(struct srcinst_string_list* dep,
 	ft != SRCINST_TYPE_DIR) {
       /* only existing files, links and dirs are considered for dependencies.
          Mark element for removal */
+      if (!mirror)
+	free(dep->strings[i]);
       dep->strings[i] = 0; n--;
 
     } else {
@@ -430,13 +403,18 @@ static int _get_relative_dependencies(struct srcinst_string_list* dep,
       
       for (this = packages->first; this; this = this->next) {
 	struct srcinst_string_list* this_list;
-	/* of course I depend on myself, so skip this->info == info */
+	/* of course I depend on myself, so skip 
+	   this->info == info */
 	if (this->info == info)
 	  continue;
-
+	
 	if (!this->info->loaded) {
 	  if (!_load_package_info(this->info)) {
-	    free(dep->strings); dep->strings = 0; return 0;
+	    if (mirror) 
+	      { free(dep->strings); dep->strings = 0; }
+	    else
+	      { srcinst_free_string_list(dep); }
+	    return 0;
 	  }
 	}
 	
@@ -451,6 +429,8 @@ static int _get_relative_dependencies(struct srcinst_string_list* dep,
       if (!found) {
 	/* there is no dependency for this file entry.
 	   Mark element for removal */
+	if (!mirror)
+	  free(dep->strings[i]);
 	dep->strings[i] = 0; n--;
       }
     }
@@ -494,6 +474,21 @@ static int _compare_strcmp(char** str1, char** str2) {
   return strcmp(*str1, *str2);
 }
 
+static void _strip_dirfile(struct srcinst_string_list* list) {
+  int i;
+  for (i = 0; i < list->count; i++) {
+    size_t len;
+    len = strlen(list->strings[i]);
+    if (len >= 9 &&
+	strcmp(list->strings[i] + (len - 9), "/info/dir") == 0) {
+      list->strings[i][0] = 0;
+    }
+  }
+
+  srcinst_purge_string_list(list, 0);
+}
+
+
 static int _internal_uninstall(struct _package_info* info) {
   struct srcinst_string_list dep; int i;
   int result = 0;
@@ -501,7 +496,7 @@ static int _internal_uninstall(struct _package_info* info) {
   /* get a list of all file and link dependencies for the current package.
      DEP is mirrored, not deep copied. Only (char**)->strings is alloc-ed,
      and only on success. */
-  if (!_get_relative_dependencies(&dep, info)) {
+  if (!_get_relative_dependencies(&dep, info, 1)) {
     return 0;
   }
 
@@ -516,9 +511,39 @@ static int _internal_uninstall(struct _package_info* info) {
     if ((ft == SRCINST_TYPE_FILE || ft == SRCINST_TYPE_LINK) &&
 	srcinst_bsearch(dep.strings, dep.count, filename,
 			&_access_string) == -1) {
-	    
-      if (!srcinst_spawn_wait(PRG_rm, "-f", filename, 0)) {
-	goto end_proc;
+
+      /* manage info files */
+      char* dname; char* bname; int remove;
+      remove = 1;
+      dname = srcinst_dirname(filename);
+      bname = srcinst_basename(dname);
+      if (strcmp(bname, "info") == 0) {
+	if (strcmp(srcinst_basename(filename), "dir") == 0) {
+	  /* _never_ _ever_ remove an info/dir file */
+	  remove = 0;
+
+	} else {
+	  char* dir;	  
+	  dir = srcinst_fnjoin(dname, "dir");
+	  if (srcinst_file_type(dir) == SRCINST_TYPE_FILE) {
+	    /* there is an info/dir file to update */
+	    char* infodir;
+	    infodir = srcinst_strjoin("--info-dir=", dname, 0);
+	    if (!srcinst_spawn_wait(PRG_install_info, 
+				    infodir, "--remove", filename, 0)) {
+	      srcinst_warning(SRCINST_ERR_ACTION,
+	     "could not cleanly remove infofile from infodir (ignoring)", 0);
+	    }
+	  }
+	  free(dir);
+	}
+      }
+
+      free(dname);
+
+      if (remove) {
+	if (!srcinst_spawn_wait(PRG_rm, "-f", filename, 0))
+	  goto end_proc;
       }
       
       *filename = '\0'; /* schedule for removal from the string list */
@@ -539,7 +564,7 @@ static int _internal_uninstall(struct _package_info* info) {
 
     if (srcinst_file_ltype(filename) == SRCINST_TYPE_DIR) {
       struct srcinst_string_list list;
-      if (!srcinst_glob_string_list(&list, filename, "*"))
+      if (!srcinst_glob_string_list(&list, filename, 0))
 	goto end_proc;
       
       if (list.count == 0) {
@@ -624,7 +649,7 @@ static int _make_uninstall(struct _package_info* info, char* makefile) {
 
 /* allocs return value string on the stack. Returns (char*)0 on failure */
 
-static char* _get_makefile(void) {
+char* _get_makefile(void) {
   char* filenames[] = { "makefile", "Makefile", "GNUmakefile", 0 };
   char* extensions[] = { "", ".unix", ".UNIX", ".Unix", ".linux", ".LINUX", ".Linux", ".cvs", ".CVS", 0 };
   
@@ -742,9 +767,10 @@ int _update_package(struct _package_list* l, struct _package_element* e) {
 
 SRCINST_ERR _action_configure(struct _package_info* info, char* prefix, struct srcinst_string_list* argv) {
   struct srcinst_string_list tmp; char* str;
-  
-  if (!_check_build_ready(info))
-    return SRCINST_ERR_INVALID;
+  SRCINST_ERR rv;
+
+  if ((rv = _check_build_ready(info)) != SRCINST_ERR_OK)
+    return rv;
   
   if (!_detect_configure())
     return SRCINST_ERR_MISSING;
@@ -781,39 +807,53 @@ SRCINST_ERR _action_configure(struct _package_info* info, char* prefix, struct s
 }
 
 SRCINST_ERR _action_rename_package(struct _package_element* e, char* newname) {
-  char* oldname; char* new_l;
+  struct _package_info new_info;
+  SRCINST_ERR rv = SRCINST_ERR_ACTION;
 
   if (strcmp(newname, e->info->name) == 0)
     return SRCINST_ERR_OK;
 
-  new_l = 0;
-    
-  if (e->info->source_location) {
-    new_l = srcinst_replace(e->info->source_location, e->info->name, newname);
-    if (!srcinst_spawn_wait(PRG_mv, e->info->source_location, new_l, 0)) {
-      free(new_l); return SRCINST_ERR_ACTION;
+  memcpy(&new_info, e->info, sizeof(struct _package_info));
+  new_info.name = newname;
+
+  if (new_info.source_location) {
+    new_info.source_location = 
+      srcinst_replace(e->info->source_location,
+		      e->info->name, new_info.name);
+  }
+
+  if (!_save_package_info(&new_info)) {
+    goto end_proc;
+  }
+
+  if (new_info.source_location) {
+    if (!srcinst_spawn_wait(PRG_mv, "-f", e->info->source_location,
+			    new_info.source_location, 0)) {
+      _unlink_package_info(&new_info);
+      goto end_proc;
     }
-    free(e->info->source_location);
-    e->info->source_location = srcinst_strdup(new_l);
-    free(new_l);
   }
-
-  oldname = e->info->name;
-  e->info->name = newname;
-
-  if (!_save_package_info(e->info)) {
-    e->info->name = oldname;		/* fallback to a safe state */
-    return SRCINST_ERR_ACTION;
-  }
-
-  e->info->name = oldname;
+      
   _unlink_package_info(e->info);	  /* remove old from the filesystem */
   
+  free(e->info->name);
   e->info->name = srcinst_strdup(newname);
+
+  if (e->info->source_location) {
+    free(e->info->source_location);
+    e->info->source_location = new_info.source_location;
+  }
+
   _detach_package_list(&_srcinst_state.packages, e);
   _add_package_element(&_srcinst_state.packages, e);
-  
+
   return SRCINST_ERR_OK;
+
+ end_proc:
+  /* note that falls back to safe state on error */
+  if (new_info.source_location)
+    free(new_info.source_location);
+  return rv;
 }
 
 int _detect_configure(void) {
@@ -822,9 +862,14 @@ int _detect_configure(void) {
     "./autogen.sh", "./autogen", "./bootstrap.sh", "./bootstrap", 0
   };
   
-  if ((ft = srcinst_file_type("./configure")) == SRCINST_TYPE_FILE)
-    return 1;
-
+  if ((ft = srcinst_file_type("./configure")) == SRCINST_TYPE_FILE) {
+    if (srcinst_file_access("./configure", "x"))
+      return 1;
+    else
+      srcinst_warning(SRCINST_ERR_PERM,
+		      "configure script is not executable", 0);
+  }
+  
   for (i = 0; autogen_scripts[i]; i++) {
     if ((ft = srcinst_file_type(autogen_scripts[i])) == SRCINST_TYPE_FILE) {
       /* found a generation script */
@@ -832,12 +877,18 @@ int _detect_configure(void) {
       executable = srcinst_file_access(autogen_scripts[i], "x");
       if (srcinst_spawn_wait(executable ? autogen_scripts[i] : PRG_sh,
 			     executable ? 0 : autogen_scripts[i], 0)) {
+
 	if ((ft = srcinst_file_type("./configure")) == SRCINST_TYPE_FILE) {
-	  return 1;
+	  if (srcinst_file_access("./configure", "x"))
+	    return 1;
+	  else 
+	    srcinst_warning(SRCINST_ERR_PERM,
+			    "configure script is not executable", 0);
 	}
       }
     }
   }
+
   return 0;
 }
 
@@ -846,7 +897,7 @@ int _detect_configure(void) {
    in directory DIRNAME or one of its subdirectories (recurs).
    Place all these files in the FILELIST string list. */
 
-static int _find_files_aux(struct srcinst_string_list* filelist, char* dirname, char** good_files) {
+int _find_files_aux(struct srcinst_string_list* filelist, char* dirname, char** good_files) {
   struct srcinst_string_list dirlist; int i, rv;
 
   for (i = 0; good_files[i]; i++) {
@@ -862,7 +913,7 @@ static int _find_files_aux(struct srcinst_string_list* filelist, char* dirname,
   rv = 1;
 
   srcinst_init_string_list(&dirlist);
-  if (!srcinst_glob_string_list(&dirlist, dirname, "*"))
+  if (!srcinst_glob_string_list(&dirlist, dirname, 0))
     return 0;
   
   for (i = 0; i < dirlist.count; i++) {
@@ -884,9 +935,10 @@ SRCINST_ERR _action_find_configure(struct _package_info* i, struct srcinst_strin
   char* good_scripts[] = { 
     "configure", "autogen.sh", "autogen", "bootstrap.sh", "bootstrap", 0
   };
+  SRCINST_ERR rv;
 
-  if (!_check_build_ready(i))
-    return SRCINST_ERR_INVALID;
+  if ((rv = _check_build_ready(i)) != SRCINST_ERR_OK)
+    return rv;
 
   srcinst_init_string_list(scripts);
   if (!_find_files_aux(scripts, ".", good_scripts)) {
@@ -900,9 +952,10 @@ SRCINST_ERR _action_find_configure(struct _package_info* i, struct srcinst_strin
 SRCINST_ERR _action_detect_autoconf(struct _package_info* i, float* v) {
   FILE* f; SRCINST_REGEX preg; SRCINST_MATCH m;
   char buffer[SRCINST_BUFSIZE]; size_t size; *v = 0.0;
+  SRCINST_ERR rv;
   
-  if (!_check_build_ready(i))
-    return SRCINST_ERR_INVALID;
+  if ((rv = _check_build_ready(i)) != SRCINST_ERR_OK)
+    return rv;
 
   if (!_detect_configure())
     return SRCINST_ERR_MISSING;
@@ -928,41 +981,20 @@ SRCINST_ERR _action_detect_autoconf(struct _package_info* i, float* v) {
   return SRCINST_ERR_OK;
 }
 
-struct _makefile_features* _action_detect_makefile_features(struct _package_info* i) {
-  struct _makefile_features* mf;
-  char* makefile;
-
-  if (!_check_build_ready(i))
-    return 0;
-
-  if (!(makefile = _get_makefile()))
-    return 0;
-
-  mf = srcinst_malloc(sizeof(struct _makefile_features));
-
-  if (!_test_makefile_features_rec(mf, makefile))
-    { free(mf); mf = 0; }
-
-  free(makefile);
-  return mf;
-}
-
 static int _make(char* makefile) {
   return srcinst_spawn_wait(PRG_make, "-f", makefile, 0);
 }
 
-static int _make_install(struct _package_info* info, char* makefile, int strip, int force) {
+
+/* test installation is only possible with DESTDIR or INSTALL_ROOT support */
+static SRCINST_ERR _make_test_install(struct _package_info* info, struct _makefile_features* mf, int strip) {
+
   char* destdir_variable, *install_string;
   int destdir_supported, install_root_supported, strip_supported;
 
-  struct _makefile_features mf;
-
-  if (!(_test_makefile_features_rec(&mf, makefile)))
-    return 0;
-  
-  destdir_supported = mf.features[SRCINST_FEATURE_DESTDIR].supported;
-  install_root_supported = mf.features[SRCINST_FEATURE_INSTALL_ROOT].supported;
-  strip_supported = mf.features[SRCINST_FEATURE_STRIP].supported;
+  destdir_supported = mf->features[SRCINST_FEATURE_DESTDIR].supported;
+  install_root_supported = mf->features[SRCINST_FEATURE_INSTALL_ROOT].supported;
+  strip_supported = mf->features[SRCINST_FEATURE_STRIP].supported;
 
   install_string = (strip_supported && strip ? "install-strip" : "install");
 
@@ -972,27 +1004,24 @@ static int _make_install(struct _package_info* info, char* makefile, int strip,
   } else if (install_root_supported) {
     destdir_variable = "INSTALL_ROOT";
 
-  } else if (force) {
-    destdir_variable = 0;
-    
   } else {
-    srcinst_warning(SRCINST_ERR_ACTION, "DESTDIR nor INSTALL_ROOT supported, and not forced", 0);
-    return 0;
+    return SRCINST_ERR_POLICY;
   }
 
-  if (destdir_variable) {
-    char* destdir_arg, *pattern; int rv;
+  if (1) {
+    char* destdir_arg, *pattern; int rv; 
     struct srcinst_string_list installed_files;
+
     destdir_arg = srcinst_strjoin(destdir_variable, "=",
 				  _srcinst_state.dirs[SRCINST_DIR_DEST], 0);
     
     /* do a test install in destdir */
-    rv = srcinst_spawn_wait(PRG_make, "-f", makefile, install_string,
+    rv = srcinst_spawn_wait(PRG_make, "-f", mf->makefile, install_string,
 			    destdir_arg, 0);
     free(destdir_arg);
-
+    
     if (!rv) {
-      return 0;
+      return SRCINST_ERR_ACTION;
     }
 
     /* fetch installed file list */
@@ -1002,10 +1031,12 @@ static int _make_install(struct _package_info* info, char* makefile, int strip,
 					 _srcinst_state.dirs[SRCINST_DIR_DEST], 0);
     free(pattern);
     if (!rv) {
-      return 0;
+      return SRCINST_ERR_ACTION;
     }
 
     srcinst_free_string_list(&(info->installed_files));
+    _strip_dirfile(&installed_files);
+    
     srcinst_mirror_string_list(&(info->installed_files), &installed_files);
     free(installed_files.strings);
 
@@ -1020,20 +1051,25 @@ static int _make_install(struct _package_info* info, char* makefile, int strip,
       info->installed_size = -1;
     }
   }
+  
+  return SRCINST_ERR_OK;
+}
+
+static SRCINST_ERR _make_install(struct _package_info* info, struct _makefile_features* mf, int strip) {
+  char* install_string;
+  int strip_supported;
+
+  strip_supported = mf->features[SRCINST_FEATURE_STRIP].supported;
+  install_string = (strip_supported && strip ? "install-strip" : "install");
 
   info->installed = 1;
   info->onclose |= SRCINST_ONCLOSE_UPDATE;
 
   /* now do the real install */
-  return srcinst_spawn_wait(PRG_make, "-f", makefile, install_string, 0);
+  return srcinst_spawn_wait(PRG_make, "-f", mf->makefile, install_string, 0) ?
+    SRCINST_ERR_OK : SRCINST_ERR_ACTION;
 }
 
-/*
-static int _make_clean(char* makefile) {
-  return srcinst_spawn_wait(PRG_make, "-f", makefile, "clean", 0);
-}
-*/
-
 static int _configure(char** cflags) {
   char* makefile;
 
@@ -1061,34 +1097,61 @@ int _clear_builddirs(void) {
   return srcinst_check_dirs();
 }
 
-int _check_build_ready(struct _package_info* info) {
+SRCINST_ERR _check_build_ready(struct _package_info* info) {
   if (info->build_ready)
-    return 1;
+    return SRCINST_ERR_OK;
 
   if (!info->source_location)
-    return 0;
+    return SRCINST_ERR_INVALID;
 
-  return _action_init_build(info, info->source_location, 0) == SRCINST_ERR_OK;
+  return _action_init_build(info, info->source_location, 0);
 }
 
-SRCINST_ERR _action_upgrade(struct _package_info* info) {
+SRCINST_ERR _action_upgrade(struct _package_info* old_info, struct _package_info* new_info) {
 
-  if (info->installed) {
-    if (!_internal_uninstall(info))
+  int i;
+
+  if (old_info->installed) {
+    if (!_internal_uninstall(old_info))
       return SRCINST_ERR_ACTION;
     
-    info->installed = 0;
-    info->installed_size = (off_t)-1;
-    srcinst_free_string_list(&(info->installed_files));
-    info->onclose |= SRCINST_ONCLOSE_UPDATE;
+    old_info->installed = 0;
+    old_info->installed_size = (off_t)-1;
+    srcinst_free_string_list(&(old_info->installed_files));
+    old_info->onclose |= SRCINST_ONCLOSE_UPDATE;
   }
   
-  if (info->source_location) {
+  if (old_info->source_location) {
     SRCINST_ERR rv;
-    if ((rv = _action_uninstall_source(info)) != SRCINST_ERR_OK)
+    if ((rv = _action_uninstall_source(old_info)) != SRCINST_ERR_OK)
       return rv;
   }
 
+  for (i = 0; i < new_info->installed_files.count; i++) {
+    char* dname; char* bname;
+
+    dname = srcinst_dirname(new_info->installed_files.strings[i]);
+    bname = srcinst_basename(dname);
+
+    if (strcmp(bname, "info") == 0) {
+      char* dir;	  
+      dir = srcinst_fnjoin(dname, "dir");
+      if (srcinst_file_type(dir) == SRCINST_TYPE_FILE) {
+	/* there is an info/dir file to update */
+	char* infodir;
+	infodir = srcinst_strjoin("--info-dir=", dname, 0);
+	if (!srcinst_spawn_wait(PRG_install_info, infodir,
+				new_info->installed_files.strings[i], 0)) {
+	  srcinst_warning(SRCINST_ERR_ACTION,
+			  "could not upgrade info dir (ignoring)", 0);
+	}
+	free(infodir);
+      }
+      free(dir);
+    }
+    free(dname);
+  }
+
   return SRCINST_ERR_OK;
 }
 
