@@ -1,4 +1,4 @@
-/* Source Installer, Copyright (c) 2005 Claudio Fontana
+/* Source Installer, Copyright (c) 2005-2006 Claudio Fontana
 
  srcinst.c - main interface to source installer library implementation
 
@@ -77,12 +77,19 @@ int srcinst_init(unsigned int flags) {
   _init_preferences(&_srcinst_state.preferences);
   _init_package_list(&_srcinst_state.packages);
   
-  if (!_clear_builddirs())
+  if (!srcinst_check_dirs()) {
+    /* check/ensure existence of user directory structure */
     goto err_proc;
-  
+  }
+
   if (!_lock_user()) {
     goto err_proc;
   }
+
+  if (!_clear_builddirs()) {
+    /* clear files in build directories from a previous interrupted run */
+    goto err_proc;
+  }
   
   if (srcinst_file_type(_srcinst_state.files[SRCINST_FILE_CFG]) ==
       SRCINST_TYPE_NONE) {
@@ -119,7 +126,7 @@ int srcinst_init(unsigned int flags) {
 
 static void _init_state(void) {
   memset(&_srcinst_state, 0, sizeof(struct _srcinst_state));
-  _srcinst_state.beat_delay = 100000;
+  _srcinst_state.beat_delay = 50000;
 }
 
 /* The SIGINT handler just accounts the number of signals detected in the
@@ -130,7 +137,7 @@ static void _init_state(void) {
 
 static void _handle_SIGINT(int s) {
   _srcinst_state.sigint_seen++;
-
+  
   if (_srcinst_state.waiting_for) {
     kill(srcinst_get_spawn_pid(_srcinst_state.waiting_for), SIGINT);
   }
@@ -509,11 +516,16 @@ int srcinst_close_package(SRCINST_PHANDLE h) {
 
 /* remove pending stuff from a package that has been upgraded. */
 
-SRCINST_ERR srcinst_upgrade_package(SRCINST_PHANDLE h) {
-  SRCINST_ERR rv; struct _package_element* e;
-  e = h;
+SRCINST_ERR srcinst_upgrade_package(SRCINST_PHANDLE oh, SRCINST_PHANDLE newh) {
+  SRCINST_ERR rv; struct _package_element* old, *new;
+  old = oh; new = newh;
 
-  SRCINST_INTR_ACTION(_action_upgrade(e->info));
+  if (!srcinst_close_package(newh))
+    return SRCINST_ERR_ACTION;
+
+  SRCINST_INTR_ACTION(_action_upgrade(old->info, new->info));
+
+  srcinst_close_package(old);
   return rv;
 }
 
@@ -579,6 +591,7 @@ void srcinst_free_configure_opt(struct srcinst_configure_sets* sets) {
 SRCINST_MFEATS srcinst_detect_makefile_features(SRCINST_PHANDLE h) {
   struct _package_element* e;
   e = h;
+  
   return (SRCINST_MFEATS)_action_detect_makefile_features(e->info);
 }
 
@@ -588,24 +601,65 @@ int srcinst_is_feature_supported(SRCINST_MFEATS mf, SRCINST_FEATURE idx) {
   return m->features[idx].supported;
 }
 
-void srcinst_free_makefile_features(SRCINST_MFEATS mf) {
-  free(mf);
+char* srcinst_get_makefile(SRCINST_MFEATS mf) {
+  struct _makefile_features* m;
+  m = mf;
+  return m->makefile;
 }
 
-SRCINST_ERR srcinst_install(SRCINST_PHANDLE h, int keep_binary, int keep_source, int strip, int force, SRCINST_COMP format) {
-  SRCINST_ERR rv = SRCINST_ERR_OK; struct _package_element* e;
+void srcinst_free_makefile_features(SRCINST_MFEATS mf) {
+  struct _makefile_features* m;
+  m = mf;
+  _free_makefile_features(m);
+}
+
+SRCINST_ERR srcinst_build(SRCINST_PHANDLE h, SRCINST_MFEATS mf) {
+  SRCINST_ERR rv;
+  struct _package_element* e;
+  struct _makefile_features* m;
+
+  e = h; m = mf;
+  SRCINST_INTR_ACTION(_action_build(e->info, m));
+  return rv;
+}
+
+SRCINST_ERR srcinst_test_install(SRCINST_PHANDLE h, SRCINST_MFEATS mf, int strip) {
+  SRCINST_ERR rv;
+  struct _package_element* e;
+  struct _makefile_features* m;
+
+  e = h; m = mf;
+  SRCINST_INTR_ACTION(_action_test_install(e->info, m, strip));
+  return rv;
+}
+
+int srcinst_get_conflict_list(SRCINST_PHANDLE h, struct srcinst_string_list* list) {
+  struct _package_element* e;
+
   e = h;
-
-  if (keep_source) {
-    SRCINST_INTR_ACTION(_action_install_source(e->info, format));
-    if (rv != SRCINST_ERR_OK)
-      return rv;
-  }
+  return _action_get_conflict_list(e->info, list);
+}
   
-  if (keep_binary) {
-    SRCINST_INTR_ACTION(_action_install_binary(e->info, strip, force));
-  }
+SRCINST_ERR srcinst_install_binary(SRCINST_PHANDLE h, SRCINST_MFEATS mf, int strip) {
+  SRCINST_ERR rv;
+  struct _package_element* e;
+  struct _makefile_features* m;
+  e = h; m = mf;
 
+  SRCINST_INTR_ACTION(_action_install_binary(e->info, m, strip));
+  return rv;
+}
+
+/* parameter MF is unused, but it's helpful because it makes it clear
+   that only configured source, with Makefiles ready, must be installed */
+
+SRCINST_ERR srcinst_install_source(SRCINST_PHANDLE h, SRCINST_MFEATS mf, SRCINST_COMP format) {
+  SRCINST_ERR rv;
+  struct _package_element* e;
+  /* struct _makefile_features* m; */
+  e = h; /* m = mf; */
+
+  SRCINST_INTR_ACTION(_action_install_source(e->info, format));
   return rv;
 }
 
@@ -799,7 +853,7 @@ int srcinst_set_descr_long(SRCINST_PHANDLE h, char* descr_long) {
   }
   e->info->long_description = descr_long ? srcinst_strdup(descr_long) : 0;
   e->info->onclose |= SRCINST_ONCLOSE_UPDATE;
-  return 0;
+  return 1;
 }
 
 char* srcinst_get_configure_string(SRCINST_PHANDLE h) {
@@ -861,3 +915,4 @@ struct srcinst_string_list* srcinst_get_installed_files(SRCINST_PHANDLE h) {
   e = h;
   return &(e->info->installed_files);
 }
+
